/*******************************************************************
 Project : SUN Legacy Type5 keyboard -> USB converter.
 Author  : IIDA Tetsushi (hogeman@gmail.com)
 Date    : 2009-01-20
 License : GNU GPL v2

 based on other projects such as "HIDKeys", "Dulcimer", etc. 
 *******************************************************************/
/*
#define F_CPU 12000000UL
*/

#include <avr/io.h>
#include <avr/pgmspace.h> 
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <string.h>

#include "usbdrv/usbdrv.h"

/*--------------------------------------------------------------*/
/*	definition about USB keyboard				*/
/*	see : "Device Class Definition for HID v.1.11"		*/ 
/*--------------------------------------------------------------*/
/* USB keyboard report descriptor (created by USB.org's tool)	*/

PROGMEM char usbHidReportDescriptor[64] = {
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x06,                    // USAGE (Keyboard)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x95, 0x08,                    //   REPORT_COUNT (8)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
    0x81, 0x02,                    //   INPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x81, 0x01,                    //   INPUT (Cnst,Ary,Abs)
    0x95, 0x05,                    //   REPORT_COUNT (5)
    0x75, 0x01,                    //   REPORT_SIZE (1)
    0x05, 0x08,                    //   USAGE_PAGE (LEDs)
    0x19, 0x01,                    //   USAGE_MINIMUM (Num Lock)
    0x29, 0x05,                    //   USAGE_MAXIMUM (Kana)
    0x91, 0x02,                    //   OUTPUT (Data,Var,Abs)
    0x95, 0x01,                    //   REPORT_COUNT (1)
    0x75, 0x03,                    //   REPORT_SIZE (3)
    0x91, 0x01,                    //   OUTPUT (Cnst,Ary,Abs)
    0x95, 0x06,                    //   REPORT_COUNT (6)
    0x75, 0x08,                    //   REPORT_SIZE (8)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
    0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
    0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
    0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
    0xc0                           // END_COLLECTION
};



/* USB keyboard report buffer */
static struct REPORT_BUF_STRUCT {
	uchar modKey;
	uchar reserved;
	uchar keys[6];
} reportBuffer;



static uchar idleRate;    		/* in 4 ms units */
static uchar protocolVer = 1;		/*  0 = boot protocol, 1 = report protocol */
static uchar expectReport;		/* flag to indicate if we expect an USB-report */
static uchar keyDidChange;		/* key status was changed */



/*--------------------------------------------------------------*/
/*	for macro command					*/
/*--------------------------------------------------------------*/
enum keyMacroState {
	KM_NORMAL = 0,			/* normal operation */
	KM_RECORDING,			/* on recording */
	KM_RUNNING			/* on running */
};

static enum keyMacroState kmStat;		/* current operation mode */
static uchar kmIdx;					/* stSeq buffer selector (0..3) */

struct STROKE_SEQUENCE_BUF {
	uchar len;			/* effective data length */
	uchar keys[63];		/* key stroke buffer */
};

static struct STROKE_SEQUENCE_BUF rom_stSeq[4] EEMEM = { 	/* 256bytes(=minimum size of ATMega?8) */
	{0, {0}}, {0, {0}}, {0, {0}}, {0, {0}}};		/* initialize by zero */
static struct STROKE_SEQUENCE_BUF stSeqOnRec;			/* key stroke Recording buffer */
static struct STROKE_SEQUENCE_BUF stSeq[4];			/* key stroke buffer */

uchar sunkbdbuffer[16];			/* A Type5c-tol kapott kod ebbe a pufferbe kerul, igy nem az IT dolgozza fel az egesz pakkot */
uchar sunkbdbuffer_wp;			/* Beiro pointer a pufferhez, ahol all, oda johet a kovetkezo kod */
uchar sunkbdbuffer_rp;			/* Olvaso pointer, a meg ki nem olvasott pozicion all. Ha egyezik a beiroval, nincs adat! */
uchar sunkbd_powerbutton;		/* POWER nyomogomb allapot */



/*--------------------------------------------------------------*/
/*	definition about SUN Keyboard (Serial)			*/
/*--------------------------------------------------------------*/
#define UBRR_Value(b)       ((F_CPU-8*b)/(16*b))

/* SUN keyboard commands */
#define SKBDCMD_RESET       0x01
#define SKBDCMD_BELLON      0x02
#define SKBDCMD_BELLOFF     0x03
#define SKBDCMD_SETLED      0x0e


/* The led set commands require sending the SETLED byte then
 * a byte encoding which led's to have set.  Here are the bit
 * values, a bit set = led-on. */
#define SKBD_LED_NLOCK          0x01   /* Num-lock */
#define SKBD_LED_CMPOSE         0x02   /* Compose */
#define SKBD_LED_SCRLCK         0x04   /* Scroll-lock */
#define SKBD_LED_CLOCK          0x08   /* Caps-lock */
#define SKBD_LED_KANA           0x10   /* kana */

/* USB equivalents. order of bits are different from SUN's definition. */
#define USB_LED_NLOCK          	0x01   /* Num-lock */
#define USB_LED_CLOCK          	0x02   /* Caps-lock */
#define USB_LED_SCRLCK        	0x04   /* Scroll-lock */
#define USB_LED_CMPOSE        	0x08   /* Compose */
#define USB_LED_KANA          	0x10   /* kana */

/* Special state characters */
#define SKBD_RESET          0xff
#define SKBD_RESET_NEXT     0x04		/* RESET kodot koveto kod, ha minden rendben van */
#define SKBD_LYOUT          0xfe
#define SKBD_ALLUP          0x7f

/* Special Keys */
#define SKBD_HELP			0x76
#define SKBD_STOP			0x01
#define SKBD_AGAIN			0x03
#define SKBD_PROPS			0x19
#define SKBD_UNDO			0x1a
#define SKBD_FRONT			0x31
#define SKBD_COPY			0x33
#define SKBD_OPEN			0x48
#define SKBD_PASTE			0x49
#define SKBD_FIND			0x5f
#define SKBD_CUT			0x61
/* #define SKBD_LINEFEED		0x6F */ /* Japanese Keyboard uses this key as normal character */
#define SKBD_POWER			0x30



/* SUN keycode -> USB Usage ID mapping. (zero == invalid)  */
const prog_uchar sun_type5_usbmap[128] = {
   0,  0,  129,   0, 128,  58,  59,  67,	/* 0x00-0x07 */
  60,  68,  61,  69,  62, 230,  63,  41,	/* 0x08-0x0f 0x0f = ESC lesz a felirat nelkuli bill is! */
  64,  65,  66, 226,  82,  72,  70,  71,	/* 0x10-0x17 */
  80,   0,   0,  81,  79,  41,  30,  31,	/* 0x18-0x1f */
  32,  33,  34,  35,  36,  37,  38,  39,	/* 0x20-0x27 */
  45,  46,  53,  42,  73, 127,  84,  85,	/* 0x28-0x2f */
   0,   0,  99,   0,  74,  43,  20,  26,	/* 0x30-0x37 */
   8,  21,  23,  28,  24,  12,  18,  19,	/* 0x38-0x3f */
  47,  48,  76, 228,  95,  96,  97,  86,	/* 0x40-0x47 0x43 = Compose 101, RCTRL 228 */
   0,   0,  77,   0, 224,   4,  22,   7,	/* 0x48-0x4f */
   9,  10,  11,  13,  14,  15,  51,  52,	/* 0x50-0x57 */
  49,  40,  88,  92,  93,  94,  98,   0,	/* 0x58-0x5f */
  75,   0,  83, 225,  29,  27,   6,  25,	/* 0x60-0x67 */
   5,  17,  16,  54,  55,  56, 229,   0,	/* 0x68-0x6f */
  89,  90,  91,   0,   0,   0,   0,  57,	/* 0x70-0x77 */
 227,  44, 231,  78,   0,  87,   0,   0 	/* 0x78-0x7f */
};



/*--------------------------------------------------------------*/
/*	Function Prototypes					*/
/*--------------------------------------------------------------*/
static void wait_ms(int);
static void hardwareInit(void);
static void uart_putchar(uchar);
static uchar sunCode2UsageId(uchar);



/*--------------------------------------------------------------*/
/*	functions from here					*/
/*--------------------------------------------------------------*/
/*
	PB0	O	Debug LED: Hi: LED ON
	PB1	I	Power Button: Hi: Power Button pressed
	PB2	x
	PB3	MOSI	ICSP
	PB4	MISO	ICSP
	PB5	SCK	ICSP
	PB6	XTAL1
	PB7	XTAL2
	PC0	x
	PC1	x
	PC2	x
	PC3	x
	PC4	x
	PC5	x
	PC6	RESET	ICSP
	PD0	I	RxD from SunKBD
	PD1	O	TxD to SunKBD
	PD2	IO	USB D+ (INT0)
	PD3	IO	USB D- (INT1)
	PD4	x
	PD5	O	USB PullUp Control
	PD6	x
	PD7	x
*/




/* Wait function: */
static void wait_ms(int m) {
  while(m--) _delay_ms(1);
}



/* IO Port initialization: */
static void hardwareInit(void) {
  PORTB = 0xff;   /* activate all pull-ups */
  DDRB = 0x01;    /* all pins input except LED */
  PORTC = 0xff;   /* activate all pull-ups */
  DDRC = 0x00;    /* all pins input */
  PORTD = 0xd3;   /* 1101 0011 bin: activate pull-ups except on USB lines */
  DDRD  = 0x20;   /* 0010 0000 bin: all pins input except USB reset */

  usbDeviceDisconnect();					/* USB disconnect -> reconnect */
  wait_ms(50);   			/* wait 50ms */		/* probably this is not required as following lines are vely slow */

  eeprom_busy_wait();						/* restore keystroke previously saved from EEPROM (slow) */
  eeprom_read_block(&stSeq, &rom_stSeq, sizeof(stSeq));		/* read saved keystroke */

  usbDeviceConnect();

  usbInit();

  TCCR0B = 4;      		/* timer 0 prescaler: 256 */	/* configure timer 0 for a rate of 16M/(256 * 256) = 244.14 Hz (4.096ms) */
	
  UBRR0H = UBRR_Value(1200) >> 8;				/* serial settings */
  UBRR0L = UBRR_Value(1200) &  0xFF;
  UCSR0B = _BV(RXEN0) | _BV(TXEN0) | _BV(RXCIE0);	
}





/* SEND command to SUN Keyboard: */
static void uart_putchar(uchar c) {
  loop_until_bit_is_set(UCSR0A, UDRE0);
  UDR0 = c;
}





/* build a USB ReportBuffer from Sun Keyboard Code: */ 
static uchar sunCode2UsageId(uchar rb) {
  uchar cnt;
  uchar keyUp = rb & 0x80;						/* 0:Down, ~0:Up */

  uchar usbKey = pgm_read_byte(&(sun_type5_usbmap[rb & 0x7f]));		/* Get USB Keycode */

  if (usbKey == 0) {							/* ignore N/A key */
    return 0;								/* not changed */
  } else if ((0xe0 <= usbKey) && (usbKey <= 0xe7)) {			/* Check Modifier Keys */
    if (keyUp) { 
      reportBuffer.modKey &=  ~(1 << (usbKey - 0xe0));
    } else {
      reportBuffer.modKey |=  (1 << (usbKey - 0xe0));
    }
  } else {								/* Check Normal Keys */		/* parse keys */
    if (keyUp) {							/* when key release */
      for (cnt = 0; cnt < sizeof reportBuffer.keys; cnt++) {		/* draw ahead required? */ 
        if (reportBuffer.keys[cnt] == usbKey) {
          reportBuffer.keys[cnt] = 0;
          break;
        }
      }
    } else {								/* when key press */
      for (cnt = 0; cnt < sizeof reportBuffer.keys; cnt++) {
        if (reportBuffer.keys[cnt] == 0) {
          reportBuffer.keys[cnt] = usbKey;
          break;
        }
      }
    }
  } 
  return 1;								/* key did change */
}





/* Adat elhozasa a SunKBD felol, ez IRQ-ban megy: */
ISR(USART_RX_vect) {
  uchar b = UDR0;			/* Adat elhozva az USART regiszterebol */
  sei();				/* IT engedve, hogy az USB ne akadjon meg */
  sunkbdbuffer_wp&= 0x0f;		/* 0..15 a puffermutato */
  sunkbdbuffer[sunkbdbuffer_wp] = b;	/* Pufferbe kapott kod mentve */
  sunkbdbuffer_wp++;			/* Kovetkezo poziciora all */
  sunkbdbuffer_wp&= 0x0f;		/* 0..15 a puffermutato */
}





/* Sun Keyboardtol jovo kodok pufferenek a feldolgozasa */
void SunKeyboard_Buffer(void) {
  static uchar pwrOnFlag;  		/* to ignore "0xff 0x04" when power on. */
  if (sunkbdbuffer_wp != sunkbdbuffer_rp) {
    sunkbdbuffer_rp &= 0x0f;				/* Olvasopointer 0..15 */
    uchar rb = sunkbdbuffer[sunkbdbuffer_rp];		/* Adat beo a pufferbol */
    sunkbdbuffer_rp++;					/* Kovetkezo elemre all */
    sunkbdbuffer_rp &= 0x0f;				/* Olvasopointer 0..15 */
    uchar keyUp = rb & 0x80;			/* 0:Down, ~0:Up */
    uchar vKey  = rb & 0x7f;

    if (kmStat == KM_RUNNING) {			/* ignore any key on macro running */
      return;
    }

    if (rb == SKBD_RESET) {			/* On Power ON ... */
      pwrOnFlag = 1;				/* ignore following character */
      return;
    }
    if (pwrOnFlag) {
      pwrOnFlag = 0;
      if (rb == SKBD_RESET_NEXT)
        return;					/* ignore keyboard type */
    }
    if (rb == SKBD_ALLUP) {			/* reset keyboard state */
      memset((void *)&reportBuffer, 0, sizeof reportBuffer);
      keyDidChange = 1;
    } else if (vKey == SKBD_HELP) {			/* HELP lenyomasa: makro felvetel / visszajatszas */
      if (!keyUp) {					/* Csak lenyomasnal */
        if (sunkbd_powerbutton != 0) {		/* POWER nyomva, felvetel jon! */
          uart_putchar(SKBDCMD_BELLON);
          memset(&stSeqOnRec, 0, sizeof stSeqOnRec);
          kmStat = KM_RECORDING;
          uart_putchar(SKBDCMD_BELLOFF);
        } else {					/* POWER nincs nyomva, felvetel vege vagy visszajatszas jon! */
          if (kmStat == KM_RECORDING) {		/* Felvetel mod volt, SAVE */
            uart_putchar(SKBDCMD_BELLON);
            kmStat = KM_NORMAL;
            stSeq[0] = stSeqOnRec;
            eeprom_busy_wait();			/* save to eeprom (very slow!) */
            eeprom_write_block(&(stSeq[0]), &(rom_stSeq[0]), sizeof(stSeq[0]));
            uart_putchar(SKBDCMD_BELLOFF);
          } else if (kmStat == KM_NORMAL) {		/* Normal mod volt, PLAY */
            kmIdx = 0; 
            kmStat = KM_RUNNING;			/* execute at main loop */
          }
        }
      }
    } else if (vKey == SKBD_CUT) {			/* Extra gombok kezelese: */
      if (keyUp) {			/* CUT felengedes */
        sunCode2UsageId(0xe5);		/* X felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* CUT lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x65);		/* X lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_PASTE) {
      if (keyUp) {			/* PASTE felengedes */
        sunCode2UsageId(0xe7);		/* V felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* PASTE lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x67);		/* V lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_COPY) {
      if (keyUp) {			/* COPY felengedes */
        sunCode2UsageId(0xe6);		/* C felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* COPY lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x66);		/* C lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_UNDO) {
      if (keyUp) {			/* UNDO felengedes */
        sunCode2UsageId(0xe4);		/* Z felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* UNDO lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x64);		/* Z lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_AGAIN) {
      if (keyUp) {			/* AGAIN felengedes */
        sunCode2UsageId(0xe4);		/* Z felengedese */
        sunCode2UsageId(0xe3);		/* LSHIFT felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* AGAIN lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x63);		/* LSHIFT lenyomasa */
        sunCode2UsageId(0x64);		/* Z lenyomasa */
        keyDidChange = 1;
      }

    } else if (vKey == SKBD_FIND) {
      if (keyUp) {			/* FIND felengedes */
        sunCode2UsageId(0xd0);		/* F felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* FIND lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x50);		/* F lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_OPEN) {
      if (keyUp) {			/* FIND felengedes */
        sunCode2UsageId(0x88);		/* F3 funkciobillentyu felengedese */
        keyDidChange = 1;
      } else {			/* FIND lenyomasa */
        sunCode2UsageId(0x08);		/* F3 funkciobillentyu lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_FRONT) {
    } else if (vKey == SKBD_PROPS) {
      if (keyUp) {			/* PROPS felengedes */
        sunCode2UsageId(0xbe);		/* O felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* PROPS lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x3e);		/* O lenyomasa */
        keyDidChange = 1;
      }
    } else if (vKey == SKBD_STOP) {
      if (keyUp) {			/* STOP felengedes */
        sunCode2UsageId(0xce);		/* S felengedese */
        sunCode2UsageId(0xcc);		/* CTRL felengedese */
        keyDidChange = 1;
      } else {			/* STOP lenyomasa */
        sunCode2UsageId(0x4c);		/* CTRL lenyomasa */
        sunCode2UsageId(0x4e);		/* S lenyomasa */
        keyDidChange = 1;
      }							/* Extra gombkezelo resz eddig. */

    } else {						/* real key event */
      if (kmStat == KM_RECORDING) {			/* record keystrokes */
        if (stSeqOnRec.len < sizeof stSeqOnRec.keys) {
          stSeqOnRec.keys[stSeqOnRec.len] = rb;
          stSeqOnRec.len++;
        } else {
          kmStat = KM_NORMAL;
          uart_putchar(SKBDCMD_BELLON);
          memset(&stSeqOnRec, 0, sizeof stSeqOnRec);		
          uart_putchar(SKBDCMD_BELLOFF);		
        }
      }

      if (sunCode2UsageId(rb)) {				/* make usb keycode */
        keyDidChange = 1;
      }
    }
  }
}





/*
 * The write function is called when LEDs should be set. 
 */
uchar usbFunctionWrite(uchar *data, uchar len) {
  uchar cLED = 0;

  if (expectReport && (len == 1)) {
    if (data[0] & USB_LED_NLOCK) {
      cLED |= SKBD_LED_NLOCK;
    }
    if (data[0] & USB_LED_CLOCK) {
      cLED |= SKBD_LED_CLOCK;
    }
    if (data[0] & USB_LED_SCRLCK) {
      cLED |= SKBD_LED_SCRLCK;
    }
    if (data[0] & USB_LED_CMPOSE) {
      cLED |= SKBD_LED_CMPOSE;
    }
    if (data[0] & USB_LED_KANA) {
      cLED |= SKBD_LED_KANA;
    }
    uart_putchar(SKBDCMD_SETLED);
    uart_putchar(cLED);
  }
  expectReport = 0;
  return 1;
}





/*
 * This function is called whenever we receive a setup request via USB.
 */
usbMsgLen_t	usbFunctionSetup(uchar data[8]) {
  usbRequest_t    *rq = (void *)data;

  usbMsgPtr = (uchar *)&reportBuffer;
  if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {	/* class request type */
    if(rq->bRequest == USBRQ_HID_GET_REPORT) {				/* wValue: ReportType (highbyte), ReportID (lowbyte) */
/* we only have one report type, so don't look at wValue */
      return sizeof(reportBuffer);
    } else if(rq->bRequest == USBRQ_HID_SET_REPORT) {
      if (rq->wLength.word == 1) {					/* We expect one byte reports */
        expectReport = 1;
        return USB_NO_MSG;						/* Call usbFunctionWrite with data */
      }
    } else if(rq->bRequest == USBRQ_HID_GET_IDLE) {
      usbMsgPtr = &idleRate;
      return 1;
    } else if(rq->bRequest == USBRQ_HID_SET_IDLE) {
      idleRate = rq->wValue.bytes[1];
    } else if (rq->bRequest == USBRQ_HID_GET_PROTOCOL) {
      if (rq->wValue.bytes[1] < 1) {
        protocolVer = rq->wValue.bytes[1];
      }
    } else if(rq->bRequest == USBRQ_HID_SET_PROTOCOL) {
      usbMsgPtr = &protocolVer;
      return 1;
    }
  } else {
/* no vendor specific requests implemented */
  }
  return 0;
}





/*
 * main function.
 */
int main(void) {
  uchar idleCounter = 0;
  uchar sendCnt = 0;

  hardwareInit();

  sunkbdbuffer_wp = 0;		/* Iro pointer init */
  sunkbdbuffer_rp = 0;		/* Olvaso pointer init */

  wdt_enable(WDTO_1S);

  sei(); 			/* enable interrupts */

  for (;;) {		/* ensure loop within 50ms! */
    wdt_reset();
    usbPoll();

    if (TIFR0 & _BV(TOV0)) {		/* 4ms period expired */ 
      TIFR0 = _BV(TOV0);
      if(idleRate != 0){
        if(idleCounter > 0){
          idleCounter--;
        } else {
          idleCounter = idleRate;
          keyDidChange = 1;
        }
      }
    }

    if ((PINB & 0x02) == 0) {				/* POWER nyomogomb allapot beolvasasa */
      sunkbd_powerbutton = 0;
    } else {
      sunkbd_powerbutton = 1;
    }

    SunKeyboard_Buffer();				/* Type5c-tol kapott kodok pufferenek feldolgozasa */

    if (kmStat == KM_RUNNING) {				/* macro mode */
      if (sendCnt < stSeq[kmIdx].len) {			/* execute macro */
        if (usbInterruptIsReady()) {
          sunCode2UsageId(stSeq[kmIdx].keys[sendCnt]);
          usbSetInterrupt((uchar *)&reportBuffer, sizeof(reportBuffer)); // send
          sendCnt++;
        }
      } else {
        sendCnt = 0;					/* exit to normal mode */
        kmStat = KM_NORMAL;
      }
    } else {
      if(keyDidChange && usbInterruptIsReady()) {
        keyDidChange = 0;
        usbSetInterrupt((uchar *)&reportBuffer, sizeof(reportBuffer));
      }
    }
  }
  return(0);
}

